(* $Id: Korst.v,v 1.20 2011-12-01 10:03:59 deraugla Exp $ *)

Require Import Utf8.
Require Import Arith.
Require Import Omega.
Require Import Sorting.
Require Streams.

Require Import Task.
Require Import Div.
Require Import Gcd.
Require Import gcd_p.

Notation "[ x , .. , y … l ]" := (cons x .. (cons y l) ..).
Notation "[ x ]" := (cons x nil).
Notation "[ ]" := nil.
Notation "x ≤ y ≤ z" := (x ≤ y ∧ y ≤ z) (at level 70, y at next level).
Notation "x < y ≤ z" := (x < y ∧ y ≤ z) (at level 70, y at next level).
Infix "++" := app (right associativity, at level 60).

Definition nth_time_unit t task :=
  if lt_dec t (phase task) then E
  else
    let t' := f_mod (t - phase task) (period task) in
    if lt_dec t' (duration task) then A (id task) else E.

Fixpoint merge_unit_list t tasks :=
  match tasks with
  | [] => Some E
  | [task … tl] =>
      match nth_time_unit t task with
      | E => merge_unit_list t tl
      | A k =>
          match merge_unit_list t tl with
          | None => None
          | Some E => Some (A k)
          | Some (A _) => None
          end
      end
  end.

Definition schedulable tasks : Prop :=
  ∀ i, merge_unit_list i tasks ≠ None.

Fixpoint make_tasks_loop n s tdl :=
  match tdl with
  | [] => []
  | [(c, T) … l] => [mkTask n s c T … make_tasks_loop (S n) (s + c) l]
  end.

Definition make_tasks tdl := make_tasks_loop 1 0 tdl.

Definition schedule tl :=
  let f :=
    cofix loop i :=
      match merge_unit_list i tl with
      | None => Streams.const E
      | Some u => Streams.Cons u (loop (S i))
      end
  in
  f 0.

(* For testing *)

Fixpoint take {A} (n : nat) (s : Streams.Stream A) : list A :=
  match n with
  | O => nil
  | S n' =>
      match s with
      | Streams.Cons tu csch => [tu … take n' csch]
      end
  end.

Eval compute in take 30 (schedule (make_tasks [(1,4),(1,6),(1,4)…[]])).

(* *)

Definition korst_formula τ₁ τ₂ :=
  let g := gcd (period τ₁) (period τ₂) in
  duration τ₁ + duration τ₂ ≤ (phase τ₂ - phase τ₁) mod g + duration τ₂ ≤ g.

Definition korst_test_two_tasks τ₁ τ₂ :=
  phase τ₂ < phase τ₁ ∧ korst_formula τ₂ τ₁ \/
  phase τ₁ < phase τ₂ ∧ korst_formula τ₁ τ₂.

Lemma running_task : ∀ t i n,
  nth_time_unit i t = A n
  → i ≥ phase t ∧ f_mod (i - phase t) (period t) < duration t.
Proof.
intros t i n occu.
unfold nth_time_unit in occu.
destruct (lt_dec i (phase t)) as [I_LT_S| I_GE_S].
 discriminate occu.

 apply not_lt in I_GE_S.
 destruct (lt_dec (f_mod (i - phase t) (period t)) (duration t)).
  split; assumption.
  discriminate occu.
Qed.

Lemma running_tasks : ∀ tl i n,
  merge_unit_list i tl = Some (A n)
  → List.Exists (fun t => nth_time_unit i t = A n) tl.
Proof.
induction tl as [| t]; intros i n occu; try discriminate occu.
simpl in occu.
remember (nth_time_unit i t) as u₁.
destruct u₁ as [| n₁].
 right; apply IHtl, occu.

 left.
 remember (merge_unit_list i tl) as u₂_opt.
 destruct u₂_opt as [u₂| ]; try discriminate occu.
 destruct u₂; try discriminate occu.
 injection occu; intros; subst; symmetry ; assumption.
Qed.

Lemma impossible_both_busy : ∀ τ₁ τ₂ n₁ n₂, ∀ i,
  0 < duration τ₁ ≤ period τ₁ → 0 < duration τ₂ ≤ period τ₂
  → phase τ₂ < phase τ₁
    → nth_time_unit i τ₁ = A n₁ → nth_time_unit i τ₂ = A n₂
      → korst_formula τ₂ τ₁ → False.
Proof.
intros τ₁ τ₂ n₁ n₂ i c₁_range c₂_range s₂_lt_s₁ busy₁ busy₂ form.
unfold korst_formula in form.
remember (gcd (period τ₂) (period τ₁)) as g.
remember ((phase τ₁ - phase τ₂) mod g) as dsmg.
destruct form as (form1, form2).
replace (dsmg + duration τ₁) with (duration τ₁ + dsmg) in form1 by
  apply plus_comm.
rewrite plus_comm in form1.
apply plus_le_reg_l in form1.
remember Heqg as T₁_eq_k₁g; clear HeqT₁_eq_k₁g.
remember Heqg as T₂_eq_k₂g; clear HeqT₂_eq_k₂g.
rewrite gcd_comm in T₁_eq_k₁g.
apply gcd_mult in T₁_eq_k₁g.
apply gcd_mult in T₂_eq_k₂g.
destruct T₁_eq_k₁g as (k₁, T₁_eq_k₁g).
destruct T₂_eq_k₂g as (k₂, T₂_eq_k₂g).
apply running_task in busy₁.
apply running_task in busy₂.
assert (period τ₁ > 0) as T₁_pos by
 (apply lt_le_trans with (m := duration τ₁); destruct c₁_range; assumption).
assert (period τ₂ > 0) as T₂_pos by
 (apply lt_le_trans with (m := duration τ₂); destruct c₂_range; assumption).
assert (0 < k₁) as k₁_pos.
 destruct k₁.
  rewrite T₁_eq_k₁g in T₁_pos; apply lt_irrefl in T₁_pos; destruct T₁_pos.

  apply lt_0_Sn.

 assert (0 < k₂) as k₂_pos.
  destruct k₂.
   rewrite T₂_eq_k₂g in T₂_pos; apply lt_irrefl in T₂_pos; destruct T₂_pos.

   apply lt_0_Sn.

  destruct busy₁ as (i_ge_s₁, m₁_lt_c₁).
  destruct busy₂ as (i_ge_s₂, m₂_lt_c₂).
  rewrite T₁_eq_k₁g in m₁_lt_c₁.
  rewrite T₂_eq_k₂g in m₂_lt_c₂.
  apply mod_kb_lt in m₁_lt_c₁; try assumption.
  apply mod_kb_lt in m₂_lt_c₂; try assumption.
  replace (i - phase τ₂) with (i - phase τ₁ + (phase τ₁ - phase τ₂)) in
   m₂_lt_c₂ by omega.
  rewrite mod_plus_lt in m₂_lt_c₂.
   remember ((i - phase τ₁) mod g) as x.
   assert (duration τ₂ + x < duration τ₂ + 0) as contrad.
    apply le_lt_trans with (m := x + dsmg).
     rewrite plus_comm; apply plus_le_compat_l; assumption.

     rewrite <- Heqdsmg in m₂_lt_c₂; rewrite plus_0_r; assumption.

    apply plus_lt_reg_l in contrad.
    apply lt_n_O in contrad; destruct contrad.

   rewrite <- Heqdsmg.
   rewrite plus_comm.
   apply lt_le_trans with (m := dsmg + duration τ₁); try assumption.
   apply plus_lt_compat_l; assumption.
Qed.

Theorem Jan_Korst : ∀ tasks : list Task,
  List.Forall (fun t => 0 < duration t ≤ period t) tasks
  → List.ForallOrdPairs korst_test_two_tasks tasks
    → schedulable tasks.
Proof.
intros tl; induction tl as [| τ₁]; intros chk korst_cond;
 unfold schedulable; intros i contrad; try discriminate contrad.
inversion_clear chk as [| x₁ x₂ c_range chk_tl].
simpl in contrad.
remember (nth_time_unit i τ₁) as u₁.
destruct u₁ as [| n₁].
 unfold schedulable, not in IHtl.
 revert contrad.
 apply IHtl; inversion korst_cond; try assumption.

 remember (merge_unit_list i tl) as v.
 destruct v as [w| ].
  destruct w as [| n₂]; try discriminate contrad; clear contrad.
  symmetry  in Hequ₁.
  symmetry  in Heqv; apply running_tasks in Heqv.
  clear IHtl.
  induction tl as [| τ₂ tl₂].
   inversion Heqv.

   inversion chk_tl as [| x₁ x₂ c₂_range chk_tl₂]; subst x₁ x₂.
   inversion_clear Heqv as [x₁ x₂ Hequ₂| ].
    inversion_clear korst_cond as [| x y k r].
    inversion_clear k as [| x y test₁₂ r₁].
    unfold korst_test_two_tasks in test₁₂.
    destruct test₁₂ as [S2_LT_S1| S1_LT_S2].
     destruct S2_LT_S1 as (S2_LT_S1, form).
     apply (impossible_both_busy τ₁ τ₂ n₁ n₂ i); try assumption.

     destruct S1_LT_S2 as (S1_LT_S2, form).
     apply (impossible_both_busy τ₂ τ₁ n₂ n₁ i); try assumption.

    apply IHtl₂; try assumption.
    apply List.FOP_cons; inversion korst_cond as [| x y k r].
     inversion k; assumption.

     inversion r; assumption.

  clear contrad.
  unfold schedulable, not in IHtl.
  symmetry  in Heqv; revert Heqv.
  apply IHtl; try assumption.
  inversion korst_cond; assumption.
Qed.

(* *)

Fixpoint c_list tasks :=
  match tasks with
  | [] => []
  | [t … tl] => [duration t … c_list tl]
  end.

Fixpoint T_list tasks :=
  match tasks with
  | [] => []
  | [t … tl] => [period t … T_list tl]
  end.

Fixpoint gcd_list (l : list nat) : nat :=
  match l with
  | [] => 0
  | [x … l] => gcd x (gcd_list l)
  end.

Fixpoint sum l :=
  match l with
  | [] => 0
  | [x … l] => x + sum l
  end.

Lemma sum_list_app : ∀ t tl,
  sum (c_list (tl ++ [t])) = sum (c_list (t :: tl)).
Proof.
intros t tl.
induction tl; try reflexivity.
simpl.
rewrite IHtl.
simpl.
apply plus_permute.
Qed.

Lemma sum_list_rev : ∀ tl, sum (c_list tl) = sum (c_list (List.rev tl)).
Proof.
intros tl.
induction tl.
 reflexivity.

 simpl.
 rewrite IHtl.
 rewrite sum_list_app.
 reflexivity.
Qed.

Lemma gcd_list_app : ∀ t tl,
  gcd_list (T_list (tl ++ [t])) = gcd_list (T_list (t :: tl)).
Proof.
intros t tl.
induction tl.
 reflexivity.

 simpl.
 rewrite gcd_assoc.
 rewrite IHtl.
 simpl.
 rewrite gcd_assoc.
 replace (gcd (period a) (period t)) with (gcd (period t) (period a)) .
  reflexivity.

  apply gcd_comm.
Qed.

Lemma gcd_list_rev : ∀ tl,
  gcd_list (T_list tl) = gcd_list (T_list (List.rev tl)).
Proof.
intros tl.
induction tl.
 reflexivity.

 simpl.
 rewrite gcd_list_app.
 simpl.
 rewrite IHtl.
 reflexivity.
Qed.

Lemma gcd_pos : ∀ t tl,
  List.Forall (fun t => 0 < duration t ≤ period t) (t :: tl)
  → 0 < gcd_list (T_list (t :: tl)).
Proof.
intros t tl chk.
remember (gcd_list (T_list (t :: tl))) as g.
destruct g.
 simpl in Heqg; symmetry in Heqg.
 apply gcd_is_zero in Heqg.
 destruct Heqg as (T_eq_0).
 inversion_clear chk as [| x₁ x₂ c_range chk₁].
 rewrite T_eq_0 in c_range.
 destruct c_range as (contrad₁, contrad₂).
 apply gt_not_le in contrad₁.
 contradiction .

 apply lt_0_Sn.
Qed.

Lemma tasks_phases : ∀ tdl n s,
  Sorted (fun τ₁ τ₂ => phase τ₂ = phase τ₁ + duration τ₁)
    (make_tasks_loop n s tdl).
Proof.
induction tdl as [| (c₁, T₁)]; intros n s.
 apply Sorted_nil.

 simpl.
 apply Sorted_cons.
  apply IHtdl.

  destruct tdl as [| (c₂, T₂)].
   apply HdRel_nil.

   simpl.
   apply HdRel_cons.
   reflexivity.
Qed.

Lemma sorted_app_rem : ∀ A, ∀ R : A → A → Prop, ∀ tl₁ tl₂,
  Sorted R (tl₁ ++ tl₂) → Sorted R tl₁.
Proof.
intros A R tl₁ tl₂ sort.
induction tl₁; constructor.
 apply IHtl₁.
 simpl in sort.
 apply Sorted_inv in sort.
 destruct sort; assumption.
 
 simpl in sort.
 apply Sorted_inv in sort.
 destruct sort as (sort₁, sort₂).
 destruct tl₁; constructor.
 simpl in sort₂.
 apply HdRel_inv in sort₂; assumption.
Qed.

Inductive ForallHTPairs {A} P : list A → Prop :=
  | FHTP_nil : ForallHTPairs P []
  | FHTP_cons : ∀ x l, P x l → ForallHTPairs P l → ForallHTPairs P (x :: l).

Lemma accum_duration_last : ∀ t t' tl,
  Sorted (fun τ₁ τ₂ : Task => phase τ₂ = phase τ₁ + duration τ₁)
     (t :: tl ++ [t'])
  → phase t' = phase t + sum (c_list (t :: tl)).
Proof.
intros t t' tl sort.
revert t t' sort.
induction tl; intros t t' sort.
 simpl in sort.
 inversion sort; subst.
 inversion H2; subst.
 simpl.
 rewrite plus_0_r.
 assumption.

 simpl.
 simpl in sort.
 inversion sort; subst.
 inversion H2; subst.
 rewrite plus_assoc.
 rewrite <- H0.
 apply IHtl.
 assumption.
Qed.

Lemma accum_duration : ∀ t tl,
  Sorted (fun τ₁ τ₂ : Task => phase τ₂ = phase τ₁ + duration τ₁) (t :: tl)
  → ForallHTPairs
      (fun t' tl' => phase t' = phase t + sum (c_list (t :: tl')))
      (List.rev tl).
Proof.
intros t tl.
remember (List.rev tl) as tl'.
symmetry  in Heqtl'.
assert (tl = List.rev tl') as H.
 subst tl'; rewrite List.rev_involutive; reflexivity.

 subst tl.
 clear Heqtl'; rename tl' into tl.
 revert t; induction tl as [| τ₁]; intros t sort.
  apply FHTP_nil.

  apply FHTP_cons.
   clear IHtl.
   simpl in sort.
   simpl.
   rewrite sum_list_rev.
   replace (duration t + sum (c_list (List.rev tl))) with
    (sum (c_list (t :: List.rev tl))) .
    apply accum_duration_last.
    assumption.

    reflexivity.

   apply IHtl.
   simpl in sort.
   rewrite List.app_comm_cons in sort.
   apply sorted_app_rem in sort.
   assumption.
Qed.

Lemma list_in_rev : forall (A : Type) (x : A) (l : list A),
  List.In x l -> List.In x (List.rev l).
Proof.
intros A x l Hin.
generalize (List.in_split x l Hin).
intros (l1, (l2, H)).
rewrite H, List.rev_app_distr.
simpl.
rewrite <- List.app_assoc.
induction (List.rev l2).
 left; constructor.
 right; assumption.
Qed.

Lemma list_in_equiv_rev : forall (A : Type) (x : A) (l : list A),
  List.In x l <-> List.In x (List.rev l).
Proof.
intros A x l; split; [ apply list_in_rev | idtac ].
rewrite <- (List.rev_involutive l).
rewrite List.rev_involutive in |- * at 1.
apply list_in_rev.
Qed.

Lemma forall_rev : ∀ T P, ∀ tl : list T,
  List.Forall P tl → List.Forall P (List.rev tl).
Proof.
intros T P tl H.
rewrite List.Forall_forall in H.
induction tl using List.rev_ind.
 apply List.Forall_nil.

 simpl.
 rewrite List.rev_unit.
 apply List.Forall_cons.
  apply H.
  apply list_in_equiv_rev.
  rewrite List.rev_unit.
  apply List.in_eq.

  apply IHtl.
  intros y H1.
  apply H.
  apply list_in_equiv_rev.
  rewrite List.rev_unit.
  apply List.in_cons, list_in_equiv_rev.
  rewrite List.rev_involutive.
  assumption.
Qed.

Lemma lt_plus_duration : ∀ t x,
  0 < duration t ≤ period t → x < x + duration t.
Proof.
intros t x H.
destruct (duration t).
 destruct H; apply lt_irrefl in H; destruct H.

 unfold lt; rewrite <- plus_n_Sm.
 apply le_n_S, le_plus_l.
Qed.

Corollary Korst_list : ∀ tdl tasks,
  tasks = make_tasks tdl
  → List.Forall (fun t => 0 < duration t ≤ period t) tasks
    → sum (c_list tasks) ≤ gcd_list (T_list tasks)
      → schedulable tasks.
Proof.
intros tdl tl mk chk sum_le_gcd.
pose proof (tasks_phases tdl 1 0) as st.
unfold make_tasks in mk.
rewrite <- mk in st.
clear tdl mk.
apply Jan_Korst; try assumption.
induction tl as [| τ₁].
 apply List.FOP_nil.

 inversion_clear chk as [| x₁ x₂ range chk₁].
 apply List.FOP_cons.
  apply accum_duration in st.
  simpl in st.
  simpl in sum_le_gcd.
  clear IHtl.
  remember (List.rev tl) as tl'.
  assert (tl = List.rev tl').
   subst tl'.
   rewrite List.rev_involutive.
   reflexivity.

   subst tl.
   clear Heqtl'.
   rename tl' into tl.
   apply forall_rev.
   rewrite <- sum_list_rev in sum_le_gcd.
   rewrite <- gcd_list_rev in sum_le_gcd.
   apply forall_rev in chk₁.
   rewrite List.rev_involutive in chk₁.
   induction tl as [| τ₂].
    apply List.Forall_nil.

    inversion_clear st as [| x₁ x₂ sτ₂ hp₂].
    inversion_clear chk₁ as [| x y range₂ chk₂].
    simpl in sum_le_gcd.
    apply List.Forall_cons.
     clear IHtl.
     unfold korst_test_two_tasks.
     right.
     split.
      rewrite sτ₂.
      apply lt_le_trans with (m := phase τ₁ + duration τ₁).
       apply lt_plus_duration; try assumption.

       rewrite plus_assoc.
       apply le_plus_l.

      unfold korst_formula.
      replace (phase τ₂ - phase τ₁) with (sum (c_list (τ₁ :: tl))) .
       rewrite gcd_assoc in sum_le_gcd.
       remember (gcd (period τ₁) (period τ₂)) as g.
       assert (duration τ₁ + (duration τ₂ + sum (c_list tl)) ≤ g) as wl_le_g.
        apply le_trans with (m := gcd g (gcd_list (T_list tl))).
         assumption.

         apply gcd_le_a.
         replace g with (gcd_list (T_list [τ₁, τ₂ … []])) .
         apply gcd_pos.
         apply List.Forall_cons; try assumption.
         apply List.Forall_cons; try assumption.
         apply List.Forall_nil.

        assert (duration τ₁ < g) as H.
         apply le_trans with
            (m := duration τ₁ + (duration τ₂ + sum (c_list tl))).
          destruct (duration τ₂).
           destruct range₂; apply lt_irrefl in H; destruct H.

           simpl; rewrite <- plus_n_Sm; apply le_n_S, le_plus_l.

          assumption.

         assert (duration τ₁ + sum (c_list tl) < g) as H1.
          apply lt_le_trans with
             (m := duration τ₁ + (duration τ₂ + sum (c_list tl))).
           destruct (duration τ₂).
            destruct range₂.
            inversion H0.

            simpl.
            rewrite <- plus_n_Sm.
            apply le_n_S.
            apply plus_le_compat_l.
            apply le_plus_r.

           assumption.

          simpl.
          remember ((duration τ₁ + sum (c_list tl)) mod g) as m.
          symmetry in Heqm; apply mod_small in Heqm; try assumption.
          subst m.
          split.
           apply plus_le_compat_r, le_plus_l.

           rewrite plus_comm,  plus_permute; assumption.

       rewrite sτ₂.
       apply plus_minus.
       reflexivity.

     apply IHtl; try assumption.
     apply le_trans with (m := duration τ₁ + (duration τ₂ + sum (c_list tl))).
      apply plus_le_compat_l, le_plus_r.

      apply
       le_trans
        with
          (m := gcd (period τ₁) (gcd (period τ₂) (gcd_list (T_list tl)))).
       assumption.

       rewrite gcd_assoc.
       replace (gcd (period τ₁) (period τ₂)) with
        (gcd (period τ₂) (period τ₁)) by apply gcd_comm.
       remember (gcd (period τ₁) (period τ₂)) as g.
       rewrite <- gcd_assoc.
       rewrite gcd_comm.
       apply gcd_le_a.
       replace (gcd (period τ₁) (gcd_list (T_list tl))) with
        (gcd_list (T_list (τ₁ :: tl))) by reflexivity.
       apply gcd_pos.
       apply List.Forall_cons; assumption.

  apply IHtl; try assumption.
   destruct tl as [| τ₂].
    apply le_n.

    remember [τ₂ … tl] as tl'.
    rename tl into tl''.
    rename tl' into tl.
    apply le_trans with (m := duration τ₁ + sum (c_list tl)).
     apply le_plus_r.

     apply le_trans with (m := gcd (period τ₁) (gcd_list (T_list tl))).
      assumption.

      rewrite gcd_comm.
      apply gcd_le_a.
      subst tl.
      simpl.
      inversion chk₁; subst.
      destruct (period τ₂).
       destruct H1.
       apply le_lt_trans with (n := duration τ₂) in H; try assumption.
       apply lt_irrefl in H; destruct H.

       remember (gcd (S n) (gcd_list (T_list tl''))) as g.
       destruct g.
        symmetry in Heqg; apply gcd_is_zero in Heqg.
        destruct Heqg.
        discriminate H.

        apply lt_0_Sn.

   inversion_clear st as [| x₁ x₂ sτ₂ hp₂].
   assumption.
Qed.

Corollary Korst_two_tasks :
  ∀ c₁ T₁ c₂ T₂,
  c₁ > 0 ∧ c₂ > 0 → c₁ ≤ T₁ ∧ c₂ ≤ T₂
  → c₁ + c₂ ≤ gcd T₁ T₂
    → schedulable [mkTask 1 0 c₁ T₁, mkTask 2 c₁ c₂ T₂ … []].
Proof.
intros c₁ T₁ c₂ T₂ (c₁_pos, c₂_pos) (c₁_le_T₁, c₂_le_T₂).
intros c₁_c₂_le_gcd.
apply Korst_list with (tdl := [(c₁, T₁), (c₂, T₂) … []]).
 unfold make_tasks.
 reflexivity.

 apply List.Forall_cons; try (split; assumption).
 apply List.Forall_cons; try (split; assumption).
 apply List.Forall_nil.

 simpl.
 rewrite plus_0_r.
 rewrite gcd_0.
 assumption.
Qed.

(* conversion contraposition *)

Lemma conflict_two_tasks : ∀ τ₁ τ₂,
  0 < duration τ₁ ≤ period τ₁
  → 0 < duration τ₂ ≤ period τ₂
    → (∀ i, merge_unit_list i [τ₁, τ₂ … []] ≠ None)
    → phase τ₁ < phase τ₂
      → ¬korst_formula τ₁ τ₂
        → False.
Proof.
intros τ₁ τ₂ range₁ range₂ contrad S1_LT_S2 H.
unfold korst_formula in H.
apply Decidable.not_and in H.
 assert (0 < period τ₁) as T₁_pos by omega.
 pose proof (bezout_one_case (period τ₁) (period τ₂) T₁_pos) as bz.
 remember (gcd (period τ₁) (period τ₂)) as g.
 destruct bz as (u, bz).
 destruct bz as (v, bz).
 remember ((phase τ₂ - phase τ₁) mod g) as r.
 remember Heqr as x; clear Heqx.
 apply f_mod__mod_p, mod__eucl_div in x.
  destruct x as (eucl, r_lt_g).
  destruct eucl as (q, eucl).
  destruct H.
   apply not_le in H.
   unfold gt in H.
   rewrite plus_comm in H.
   replace (duration τ₁ + duration τ₂) with (duration τ₂ + duration τ₁) in H
    by apply plus_comm.
   apply plus_lt_reg_l in H.
   remember (phase τ₂ + v * q * period τ₂) as i.
   assert (merge_unit_list i [τ₁, τ₂ … []] = None) as conflict.
    simpl.
    remember (nth_time_unit i τ₁) as u₁.
    destruct u₁.
     assert (i = phase τ₁ + u * q * period τ₁ + r) as Hi.
      subst i.
      replace (u * q * period τ₁) with (period τ₁ * u * q) .
       rewrite bz.
       rewrite mult_plus_distr_r.
       replace (v * q * period τ₂) with (period τ₂ * v * q) .
        remember (period τ₂ * v * q) as x.
        remember (g * q) as y.
        omega.

        rewrite <- mult_assoc; apply mult_comm.

       rewrite <- mult_assoc; apply mult_comm.

      rewrite Hi in Hequ₁.
      unfold nth_time_unit in Hequ₁.
      assert (phase τ₁ ≤ phase τ₁ + u * q * period τ₁ + r) as e.
       rewrite <- plus_assoc.
       apply le_plus_l.

       destruct (lt_dec (phase τ₁ + u * q * period τ₁ + r) (phase τ₁))
        as [f| f].
        apply gt_not_le in f; contradiction .

        clear f.
        rewrite <- plus_assoc in Hequ₁.
        rewrite minus_plus in Hequ₁.
        remember ((u * q * period τ₁ + r) mod period τ₁) as x.
        rewrite plus_comm in Heqx.
        apply f_mod__mod_p, mod_plus_kb__mod in Heqx.
        apply mod_le_a in Heqx.
        apply le_lt_trans with (p := duration τ₁) in Heqx; try assumption.
        destruct (lt_dec x (duration τ₁)).
         discriminate Hequ₁.

         contradiction .

     remember (nth_time_unit i τ₂) as u₂.
     destruct u₂.
      rewrite Heqi in Hequ₂.
      unfold nth_time_unit in Hequ₂.
      assert (phase τ₂ ≤ phase τ₂ + v * q * period τ₂) as e.
       apply le_plus_l.

       destruct (lt_dec (phase τ₂ + v * q * period τ₂) (phase τ₂)) as [f| f].
        apply gt_not_le in f; contradiction .

        clear f.
        rewrite minus_plus in Hequ₂.
        remember ((v * q * period τ₂) mod period τ₂) as x.
        replace (v * q * period τ₂) with (0 + v * q * period τ₂) in Heqx .
         apply f_mod__mod_p, mod_plus_kb__mod in Heqx.
         apply mod_le_a in Heqx.
         apply le_lt_trans with (p := duration τ₂) in Heqx; try assumption.
          destruct (lt_dec x (duration τ₂)).
           discriminate Hequ₂.

           contradiction .

          destruct range₂; assumption.

         reflexivity.

      reflexivity.

    apply contrad with (i := i); assumption.

   apply not_le in H.
   remember (phase τ₁ + u * (q + 1) * period τ₁) as i.
   assert (merge_unit_list i [τ₁, τ₂ … []] = None) as conflict.
    simpl.
    remember (nth_time_unit i τ₁) as u₁.
    destruct u₁.
     rewrite Heqi in Hequ₁.
     unfold nth_time_unit in Hequ₁.
     destruct (lt_dec (phase τ₁ + u * (q + 1) * period τ₁) (phase τ₁)).
      apply gt_not_le in l.
      exfalso; apply l; apply le_plus_l.

      rewrite minus_plus in Hequ₁.
      remember ((u * (q + 1) * period τ₁) mod period τ₁) as x.
      replace (u * (q + 1) * period τ₁) with (0 + u * (q + 1) * period τ₁) in
       Heqx .
       apply f_mod__mod_p, mod_plus_kb__mod in Heqx.
       apply mod_le_a in Heqx.
       apply le_lt_trans with (p := duration τ₁) in Heqx.
        destruct (lt_dec x (duration τ₁)).
         discriminate Hequ₁.

         contradiction .

        destruct range₁; assumption.

       reflexivity.

     remember (nth_time_unit i τ₂) as u₂.
     remember (g - r) as gmr.
     destruct u₂.
      assert (i = phase τ₂ + v * (q + 1) * period τ₂ + gmr) as Hi.
       subst i.
       replace (u * (q + 1) * period τ₁) with (period τ₁ * u * (q + 1)) .
        rewrite bz.
        rewrite mult_plus_distr_r.
        replace (v * (q + 1) * period τ₂) with (period τ₂ * v * (q + 1)) .
         remember (period τ₂ * v * (q + 1)) as x.
         rewrite mult_plus_distr_l.
         rewrite mult_1_r.
         remember (g * q) as y.
         omega.

         rewrite <- mult_assoc; apply mult_comm.

        rewrite <- mult_assoc; apply mult_comm.

       rewrite Hi in Hequ₂.
       unfold nth_time_unit in Hequ₂.
       destruct
        (lt_dec (phase τ₂ + v * (q + 1) * period τ₂ + gmr) (phase τ₂)).
        apply gt_not_le in l.
        exfalso; apply l; rewrite <- plus_assoc; apply le_plus_l.

        rewrite <- plus_assoc,  minus_plus in Hequ₂.
        remember ((v * (q + 1) * period τ₂ + gmr) mod period τ₂) as x.
        rewrite plus_comm in Heqx.
        apply f_mod__mod_p, mod_plus_kb__mod in Heqx.
        apply mod_le_a in Heqx.
        apply le_lt_trans with (p := duration τ₂) in Heqx.
         destruct (lt_dec x (duration τ₂)).
          discriminate Hequ₂.

          contradiction .

         omega.

      reflexivity.

    exfalso; apply (contrad i); assumption.

  destruct g.
   apply f_gcd__gcd_p, gcd_p_is_zero in Heqg.
   destruct range₁, Heqg; omega.

   apply lt_0_Sn.

 unfold Decidable.decidable.
 omega.
Qed.

Inductive ExistsOrdPairs {T} P : list T → Prop :=
  | EOP_cons_hd : ∀ x l, List.Exists (P x) l → ExistsOrdPairs P (x :: l)
  | EOP_cons_tl : ∀ x l, ExistsOrdPairs P l → ExistsOrdPairs P (x :: l).

Theorem Jan_Korst_conversion_contraposition : ∀ tasks : list Task,
  List.Forall (fun t => 0 < duration t ≤ period t) tasks
  → ExistsOrdPairs (fun τ₁ τ₂ => not (korst_test_two_tasks τ₁ τ₂)) tasks
    → not (schedulable tasks).
Proof.
unfold schedulable.
intros tl chk ex contrad.
induction tl as [| τ₁].
 inversion ex.

 inversion_clear chk as [| x y range₁ chk₁].
 inversion_clear ex.
  clear IHtl.
  induction tl as [| τ₂].
   inversion H.

   inversion_clear chk₁ as [| x y range₂ chk₂].
   inversion_clear H.
    clear IHtl.
    unfold korst_test_two_tasks in H0.
    apply Decidable.not_or in H0.
    destruct H0 as (H1, H2).
    rewrite and_comm in H1.
    apply Decidable.not_and in H1.
     rewrite and_comm in H2.
     apply Decidable.not_and in H2.
      pose proof (lt_eq_lt_dec (phase τ₁) (phase τ₂)) as x.
      destruct x as [[S1_LT_S2| S1_EQ_S2]| S2_LT_S1].
       destruct H2; try contradiction .
       apply (conflict_two_tasks τ₁ τ₂); try assumption.
       intros i contrad2.
       simpl in contrad2.
       remember (nth_time_unit i τ₁) as u₁.
       destruct u₁.
        destruct (nth_time_unit i τ₂); discriminate contrad2.

        remember (nth_time_unit i τ₂) as u₂.
        destruct u₂.
         discriminate contrad2.

         exfalso; apply (contrad i).
         simpl.
         rewrite <- Hequ₁.
         rewrite <- Hequ₂.
         destruct (merge_unit_list i tl); try destruct t; reflexivity.

       rewrite S1_EQ_S2 in H1.
       rewrite S1_EQ_S2 in H2.
       exfalso; apply (contrad (phase τ₁)).
       simpl.
       remember (nth_time_unit (phase τ₁) τ₁) as u₁.
       destruct u₁.
        unfold nth_time_unit in Hequ₁.
        destruct (lt_dec (phase τ₁) (phase τ₁)) in Hequ₁.
         apply lt_irrefl in l; destruct l.

         rewrite minus_diag in Hequ₁.
         remember (0 mod period τ₁) as m.
         symmetry in Heqm.
         apply mod_small in Heqm.
          subst m.
          destruct (lt_dec 0 (duration τ₁)).
           discriminate Hequ₁.

           destruct range₁; contradiction .

          omega.

        rewrite S1_EQ_S2.
        remember (nth_time_unit (phase τ₂) τ₂) as u₂.
        destruct u₂.
         unfold nth_time_unit in Hequ₂.
         destruct (lt_dec (phase τ₂) (phase τ₂)) in Hequ₂.
          apply lt_irrefl in l; destruct l.

          rewrite minus_diag in Hequ₂.
          remember (0 mod period τ₂) as m.
          symmetry in Heqm.
          apply mod_small in Heqm.
           subst m.
           destruct (lt_dec 0 (duration τ₂)).
            discriminate Hequ₂.

            destruct range₂; contradiction .

           omega.

         destruct (merge_unit_list (phase τ₂) tl); try destruct t;
          reflexivity.

       destruct H1; try contradiction .
       apply (conflict_two_tasks τ₂ τ₁); try assumption.
       intros i contrad2.
       simpl in contrad2.
       remember (nth_time_unit i τ₁) as u₁.
       destruct u₁.
        destruct (nth_time_unit i τ₂); discriminate contrad2.

        remember (nth_time_unit i τ₂) as u₂.
        destruct u₂.
         discriminate contrad2.

         exfalso; apply (contrad i).
         simpl.
         rewrite <- Hequ₁.
         rewrite <- Hequ₂.
         destruct (merge_unit_list i tl); try destruct t; reflexivity.

      unfold Decidable.decidable, korst_formula; omega.

     unfold Decidable.decidable, korst_formula; omega.

    apply IHtl; try assumption.
    intros i contrad2.
    inversion contrad2.
    remember (nth_time_unit i τ₁) as u.
    destruct u.
     unfold not in contrad.
     apply contrad with (i := i).
     simpl.
     rewrite H1.
     rewrite <- Hequ.
     destruct (nth_time_unit i τ₂); reflexivity.

     remember (merge_unit_list i tl) as v.
     destruct v as [v| ].
      destruct v.
       discriminate H1.

       clear H1.
       unfold not in contrad; apply contrad with (i := i).
       simpl.
       rewrite <- Hequ.
       rewrite <- Heqv.
       remember (nth_time_unit i τ₂) as w.
       destruct w; reflexivity.

      unfold not in contrad; apply contrad with (i := i).
      simpl.
      rewrite <- Hequ.
      rewrite <- Heqv.
      destruct (nth_time_unit i τ₂); reflexivity.

  apply IHtl; try assumption.
  intros i contrad2.
  pose proof (contrad i) as contrad3.
  apply contrad3.
  simpl.
  rewrite contrad2.
  destruct (nth_time_unit i τ₁); reflexivity.
Qed.

(* conversion *)

Theorem korst_test_two_tasks_dec : ∀ τ₁ τ₂,
  Decidable.decidable (korst_test_two_tasks τ₁ τ₂).
Proof.
intros τ₁ τ₂.
unfold Decidable.decidable.
unfold korst_test_two_tasks.
unfold korst_formula.
omega.
Qed.

Theorem Forall_dec : ∀ T P, ∀ l : list T,
  (∀ x, Decidable.decidable (P x)) → Decidable.decidable (List.Forall P l).
Proof.
unfold Decidable.decidable.
intros T P l P_dec.
induction l.
 left.
 apply List.Forall_nil.

 pose proof (P_dec a) as x.
 destruct x as [P_A| NOT_P_A].
  destruct IHl as [ALL|NOT_ALL].
   left; apply List.Forall_cons; assumption.

   right; intros contrad; apply NOT_ALL; inversion contrad; assumption.

  right; intros contrad; apply NOT_P_A; inversion contrad; assumption.
Qed.

Theorem ForallOrdPairs_dec : ∀ T P, ∀ l : list T,
  (∀ x y, Decidable.decidable (P x y))
  → Decidable.decidable (List.ForallOrdPairs P l).
Proof.
unfold Decidable.decidable.
intros T P l P_dec.
induction l as [| a].
 left; apply List.FOP_nil.

 destruct IHl as [ALLP| NOT_ALLP].
  pose proof (Forall_dec T (P a) l (P_dec a)) as x.
  destruct x as [ALL| NOT_ALL].
   left; apply List.FOP_cons; assumption.

   right; intros contrad; apply NOT_ALL; inversion contrad; assumption.

  right; intros contrad; apply NOT_ALLP; inversion contrad; assumption.
Qed.

Theorem Jan_Korst_conversion : ∀ tasks : list Task,
  List.Forall (fun t => 0 < duration t ≤ period t) tasks
  → schedulable tasks
    → List.ForallOrdPairs korst_test_two_tasks tasks.
Proof.
intros tl chk.
apply Decidable.contrapositive.
 apply ForallOrdPairs_dec.
 apply korst_test_two_tasks_dec.

 intros not_korst.
 apply Jan_Korst_conversion_contraposition; try assumption.
 induction tl as [| τ₁].
  exfalso; apply not_korst, List.FOP_nil.

  inversion_clear chk as [| x y range₁ chk₁].
  pose proof
   (Forall_dec Task (korst_test_two_tasks τ₁) tl
      (korst_test_two_tasks_dec τ₁)) as x.
  destruct x as [ALL1| NOT_ALL1].
   apply EOP_cons_tl.
   apply IHtl; try assumption.
   intros; apply not_korst, List.FOP_cons; assumption.

   apply EOP_cons_hd.
   clear IHtl not_korst.
   induction tl as [| τ₂].
    exfalso; apply NOT_ALL1, List.Forall_nil.

    pose proof (korst_test_two_tasks_dec τ₁ τ₂) as x.
    destruct x as [KORST| NOT_KORST].
     apply List.Exists_cons_tl.
     apply IHtl.
      inversion chk₁; assumption.

      intros contrad.
      apply NOT_ALL1.
      apply List.Forall_cons; assumption.

     apply List.Exists_cons_hd.
     assumption.
Qed.

Corollary Korst_two_tasks_conversion :
  ∀ c₁ T₁ c₂ T₂,
  c₁ > 0 ∧ c₂ > 0 → c₁ ≤ T₁ ∧ c₂ ≤ T₂
  → schedulable [mkTask 1 0 c₁ T₁, mkTask 2 c₁ c₂ T₂ … nil]
    → c₁ + c₂ ≤ gcd T₁ T₂.
Proof.
intros c₁ T₁ c₂ T₂ (c₁_pos, c₂_pos).
intros (c₁_le_T₁, c₂_le_T₂) sched.
apply Jan_Korst_conversion in sched.
 inversion_clear sched.
 inversion_clear H.
 unfold korst_test_two_tasks in H1.
 simpl in H1; destruct H1.
  destruct H; omega.

  destruct H.
  unfold korst_formula in H1.
  simpl in H1; omega.

 apply List.Forall_cons; simpl; try split; try assumption.
 apply List.Forall_cons; simpl; try split; try assumption.
 apply List.Forall_nil.
Qed.
